// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	epicbscalgen.cc	--- Compute Boresight misalignment angles for EPIC
//
//	DESCRIPTION:
//		This is an interactive task to compute the boresight misalignment
//		angles for the instruments EMOS1/EMOS2 and EPN from observation
//		data. The Nelder/Mead simplex method is used for minimizing
//		the sum of squares of the differences between a list of
//		catalogued and observed sources positions.
//
//	CHANGE HISTORY:
//		UL, 1999-11-30: 1.0.0: - first version
//		UL, 1999-12-16: 1.1.0: - changed optimization strategy: optimize first
//								 phi/theta up to a tolerance 1.e-3; then
//								 optimize all three angles simulateneously;
//								 this avoid problems with apparent local
//								 minima in psi-parameter direction
//		UL, 1999-01-11: 1.2.0: - name of two parameters have changed:
//								   sourceRadius -> sourceradius
//								   maxIter		-> maxiter
//		UL, 2000-05-03: 1.3.0: - take declination in sum of square calculation
//								 into account
//		UL, 2003-05-23: 1.4.0: - fixed bug in
//								 SourceCorrelator::angularSeparation()
//								 which caused wrong identification in case
//								 of source confusion
//							   - new BS misalignment angles are now passed
//								 to CAL with CCF replacement data set
//								 and SAS_CCFFILES
//							   - more diagnostic output to facilitate
//								 monitoring of minimization progress
//							   - new reference source input mode in which
//								 true celestial coordinates are read from
//								 existing eboxdetect source list
//
//	AUTHOR:
//		U. Lammers, Tue Nov 30 14:34:52 MET 1999
//
//	REFERENCES:
//		[1]: XMM Star Tracker/Instrument Alignment Calibration - Interface
//			 Control Document, Issue 1.2, R. Mugellesi, ESOC/ESA
//			 Online: ftp://astro.estec.esa.nl/pub/XMM/documents/
//					 siam_icd.pdf
//
//	$Header: /sasbuild/sas_git_repo/cvs_copy/sasdev/sas/packages/ccftools/bscalgen/src/epicbscalgen.cc,v 1.10 2003/05/23 16:26:14 ulammers Exp $
//
#include <iostream>
#include <iomanip>
#include <algorithm>

#include "param.h"
#include "SetTableColumn.h"
#include "SourceList.h"
#include "Celestial.h"
#include "Numeric.h"
#include "SasTasks.h"
#include "Ds9.h"
#include "Region.h"
#include "BoresightCcf.h"
#include "CalServer.h"

#include "epicbscalgen.h"

//
//	routine to minimize
//
class SourceCorrelator {
public:
	SourceCorrelator(SourceList *slRef, Xmm::Instrument instrument,
					 bool runShowDetect = true)
		{ _slRef = slRef; _maxIter = 1000; _instrument = instrument,
			    _runShowDetect = runShowDetect; }

	static void maxIter(unsigned max) { _maxIter = max; }
	static void phi(double phi) { _phi = phi; }
	static double phi() { return _phi; }
	static void theta(double theta) { _theta = theta; }
	static double theta() { return _theta; }
	static void psi(double psi) { _psi = psi; }
	static double psi() { return _psi; }


	static double optimizePhiTheta(double angles[])
	{
		_phi = angles[1];
		_theta = angles[2];
		_psi = 0.;
		EulerAngles	eulerAngles(_phi, _theta, _psi, EulerAngles::Euler321);

		return angularSeparation(eulerAngles);
	} // optimizeThetaPhi()


	static double optimizePsi(double angle)
	{
		_psi = angle;
		
		EulerAngles	eulerAngles(_phi, _theta, _psi, EulerAngles::Euler321);

		return angularSeparation(eulerAngles);
	} // optimizePsi()


	static double optimizePhiThetaPsi(double angles[])
	{
		_phi = angles[1];
		_theta = angles[2];
		_psi = angles[3];
		EulerAngles	eulerAngles(_phi, _theta, _psi, EulerAngles::Euler321);

		return angularSeparation(eulerAngles);
	} // optimizeThetaPhi()


	static double angularSeparation(const EulerAngles &eulerAngles)
	{
		//	get new source list with for new misalignment angles
		errstr<<"New set of Euler angles in iteration: "<<_iter++<<" "
			  <<eulerAngles<<endl
			  <<Sas::message(Msg::App, Msg::Always);

		//	create replacement Boresight CCF
		const string bsCcfName("epicsbcalgenBoresight.ccf");
		BoresightCcf bsccf(bsCcfName);
		bsccf.createAndFinalize(_instrument, eulerAngles);
		//	force cal to flush all buffered atoms to ensure that new
		//	BS angles are read
		calServer.atomServer()->restart();

		errstr<<"New Boresight CCF written to `"<<bsCcfName<<"'"
			  <<Sas::message(Msg::App, Msg::Always);

		SourceList sl(_slRef->eventListName(),
					  _slRef->xAxis(), _slRef->yAxis(),
					  bsCcfName, true,
					  _slRef->binX(), _slRef->binY(), _slRef->sourceRadius(),
					  _slRef->nruns(), _slRef->likemin(), _runShowDetect);

		dataSetServer->remove(bsCcfName);
		errstr<<"New source list:\n";
		errstr<<sl<<endl
			  <<Sas::message(Msg::App, Msg::Silent);

		unsigned N = _slRef->size();
		double sumsq = 0.;
		Component selectedRefSources;
		vector<unsigned> foundRefSources;
		for (unsigned n=0; n<N; ++n) {
			//	find all sources whose counts do not differ by more than +-5%
			const Source &refSrc = (*_slRef)[n];
			double refCounts = refSrc.counts();
			errstr<<"Looking for ref-source "<<n<<" with "<<refCounts
				  <<" counts ...\n"
				  <<Sas::message(Msg::App, Msg::Always);
			vector<unsigned> indx;
			unsigned M=sl.size();
			for (unsigned m=0; m<M; ++m)
				if (sl[m].counts() >= .95 * refCounts &&
					sl[m].counts() <= 1.05 * refCounts &&
					std::find(foundRefSources.begin(),
						  foundRefSources.end(), m)	== foundRefSources.end()) {
					indx.push_back(m);
					errstr<<"Found: "<<m<<' '<<sl[m]<<endl
						  <<Sas::message(Msg::App, Msg::Always);
				}
			if (indx.size() == 0)
				errstr<<"Could not find reference source #"<<n
					  <<' '<<refSrc<<" in source list."
					  <<Sas::error("SourceMismatch");
			unsigned sn = indx[0];
			unsigned indxSize = indx.size();
			if (indxSize > 1) {
				errstr<<"There are "<<indx.size()<<" sources which have the "
						"same number of counts within +-5% - choosing the one "
						"which is closest"
					  <<Sas::warning("PossibleSourceConfusion");
				double s = 1.e20;
				for (unsigned k=0; k<indxSize; ++k) {
					double delta = refSrc.angularSeparation(sl[indx[k]]);
					if (delta < s) {
						errstr<<"Selecting provisionally src index: "
							  <<indx[k]<<' '
							  <<sl[indx[k]]<<endl
							  <<Sas::message(Msg::App, Msg::Always);
						sn = indx[k];
						s = delta;
					}
				}
			}
			foundRefSources.push_back(sn);
			errstr<<"Final selection: "<<sn<<' '<<sl[sn]<<endl
				  <<Sas::message(Msg::App, Msg::Always);
			double deltaRa = cos(refSrc.dec())*(refSrc.ra() - sl[sn].ra()) *
								RAD_TO_DEG * 3600.,
				   deltaDec = (refSrc.dec() - sl[sn].dec()) * RAD_TO_DEG*3600.;
			sumsq += deltaRa*deltaRa + deltaDec*deltaDec;
			sl[sn].shape()->color(ShapeBasics::YELLOW);
			selectedRefSources.push_back(sl[sn].shape());
		}

		//	display selected sources
        ::GeoRegion selSrcReg;
        selSrcReg.push_back(selectedRefSources);
		if (!_ds9)
			_ds9 = new Ds9;
		_ds9->displayRegion(ImageDisplayer::Region(selSrcReg,
												 ImageDisplayer::FK5));
		errstr<<"Iteration "<<_iter-1<<" - sum of squares: "
			  <<setprecision(5)<<sumsq<<" arcsec - mean angular separation: "
			  <<setprecision(3)<<sqrt(sumsq/DBL(N))<<" arcsec\n"
			  <<Sas::message(Msg::App, Msg::Always);
		if (_iter>=_maxIter)
			errstr<<"Maximum number of iterations ("<<_maxIter<<") reached"
				  <<Sas::error("MaxIterationsReached");

		return sumsq;
	} // angularSeparation()

	~SourceCorrelator() {}

private:
	static SourceList *_slRef;
	static Xmm::Instrument _instrument;
	static unsigned	_iter;
	static unsigned _maxIter;
	static double	_phi, _theta, _psi;
	static bool		_runShowDetect;
	static Ds9		*_ds9;
};

SourceList *SourceCorrelator::_slRef = 0;
unsigned SourceCorrelator::_maxIter = 0;
unsigned SourceCorrelator::_iter = 0;
double SourceCorrelator::_phi = 0;
double SourceCorrelator::_theta = 0;
double SourceCorrelator::_psi = 0;
bool SourceCorrelator::_runShowDetect = true;
Ds9 *SourceCorrelator::_ds9 = 0;
Xmm::Instrument SourceCorrelator::_instrument = Xmm::EMOS1;

void epicbscalgen()
{
	string eventTable(stringParameter("eventtable"));

	double phi0		= (realParameter("phi") + 180.*3600.)/3600. * DEG_TO_RAD,
		   theta0	= realParameter("theta")/3600. * DEG_TO_RAD,
		   psi0		= realParameter("psi") * DEG_TO_RAD;
	bool runShowDetect = booleanParameter("runshowdetect");
	double tol1 = realParameter("tol1"),
		   tol2 = realParameter("tol2");
	int	   maxIter  = intParameter("maxiter");

	SetBlock		evl(eventTable);
	Xmm::Instrument instrument;
    if (evl.table()->hasAttribute("INSTRUME"))
		instrument = Xmm::toInstrument(
						evl.table()->attribute("INSTRUME")->asString());
	else
		instrument = Xmm::toInstrument(
						evl.table()->parent()->attribute("INSTRUME")->
							asString());
	dataSetServer->close(evl.dataSet());

	string xCol = stringParameter("xcolumn"),
		   yCol = stringParameter("ycolumn");
	unsigned binx = intParameter("binx"),
			 biny = intParameter("biny"),
			 nruns = intParameter("nruns");
	double sourceRadius = realParameter("sourceradius"),
		   likemin = realParameter("likemin");

	SourceList *slRef =
		!booleanParameter("withrefsrclist")
			? new SourceList(eventTable, xCol, yCol, binx, biny,
							 sourceRadius, nruns, likemin)
			: new SourceList(eventTable, stringParameter("refsrclist"),
							 xCol, yCol, binx, biny, sourceRadius,
							 nruns, likemin);

	SourceCorrelator sc(slRef, instrument, runShowDetect);
	SourceCorrelator::phi(phi0);
	SourceCorrelator::theta(theta0);
	SourceCorrelator::psi(psi0);	

	double phi = phi0,
		   theta = theta0,
		   psi = psi0;

	errstr<<"\n---------- Boresight Angle Computation: Optimization of "
			"phi/theta ----------\n"
		  <<Sas::message(Msg::App, Msg::Always);

	int	   nVars1 = 2;
	SourceCorrelator::maxIter(maxIter);

	//	define the initial simplex
	double **p1 = Numeric::numMatrix(1, nVars1+1, 1, nVars1);

	//	first simplex vertex
	p1[1][1] = phi0;
	p1[1][2] = theta0;

	//	second simplex vertex
    p1[2][1] = phi0+50./3600.*DEG_TO_RAD;	//	phi		+50" off
	p1[2][2] = theta0+50./3600.*DEG_TO_RAD;	//	theta	+50" off

	//	third simplex vertex
	p1[3][1] = phi0-50./3600.*DEG_TO_RAD;	//	phi		-50" off
	p1[3][2] = theta0+50./3600.*DEG_TO_RAD;	//	theta	+50" off

	//	compute function for initial simplex vertices
	double *y1 = Numeric::numVector(1, nVars1+1);
	for (int n=1; n<=nVars1+1; ++n)
		y1[n] = SourceCorrelator::optimizePhiTheta(p1[n]);

	//	kick off minimization
	int nfunc1 = 0;
	Numeric::simplex(p1, y1, nVars1, tol1,
					 SourceCorrelator::optimizePhiTheta, &nfunc1);


	//
	//	stage II: optimization of phi/theta/psi
	//
	errstr<<"\n---------- Boresight Angle Computation: Optimization of "
			"phi/theta/psi ----------\n"
		  <<Sas::message(Msg::App, Msg::Always);

	int	   nVars2 = 3;
	SourceCorrelator::maxIter(maxIter);

	//	define the initial simplex
	double **p2 = Numeric::numMatrix(1, nVars2+1, 1, nVars2);

	//	first simplex vertex
	p2[1][1] = p1[1][1];
	p2[1][2] = p1[1][2];
	p2[1][3] = psi0;

	//	second simplex vertex
    p2[2][1] = p1[2][1];
	p2[2][2] = p1[2][2];
	p2[2][3] = psi0+.5*DEG_TO_RAD;				//	psi		+1/2 deg off

	//	third simplex vertex
	p2[3][1] = p1[3][1];
	p2[3][2] = p1[3][2];
	p2[3][3] = psi0-.5*DEG_TO_RAD;				//	psi		-1/2 deg off

	//	fourth simplex
	p2[4][1] = p1[1][1] + 5./3600.*DEG_TO_RAD;	//	phi		+5" off
	p2[4][2] = p1[1][2] - 5./3600.*DEG_TO_RAD;	//	theta	-5" off
	p2[4][3] = psi0-.3*DEG_TO_RAD;				//	psi		-1/3 deg off

	//	compute function for initial simplex vertices
	double *y2 = Numeric::numVector(1, nVars2+1);
	for (int n=1; n<=nVars2+1; ++n)
		y2[n] = SourceCorrelator::optimizePhiThetaPsi(p2[n]);

	//	kick off minimization
	int nfunc2 = 0;
	Numeric::simplex(p2, y2, nVars2, tol2,
					 SourceCorrelator::optimizePhiThetaPsi, &nfunc2);

	//	if we arrive hear we have a solution - optmized phi/theta are
	//	center of gravity of simplex
	phi = (p2[1][1]+p2[2][1]+p2[3][1]+p2[4][1])/4.,
	theta = (p2[1][2]+p2[2][2]+p2[3][2]+p2[4][2])/4.,
	psi = (p2[1][3]+p2[2][3]+p2[3][3]+p2[4][3])/4.;

	EulerAngles	angles(phi, theta, psi, EulerAngles::Euler321);
	const string bsCcfName("epicbscalgenBoresight.ccf");
	BoresightCcf bsccf(bsCcfName);
	bsccf.createAndFinalize(instrument, angles);

	errstr<<"Making a last iteration ...\n"
		  <<Sas::message(Msg::App, Msg::Always);
	SourceList sl(slRef->eventListName(),
				  slRef->xAxis(), slRef->yAxis(), bsCcfName, true,
				  slRef->binX(), slRef->binY(), slRef->sourceRadius());
	errstr<<sl<<endl<<Sas::message(Msg::App, Msg::Always);

	errstr<<"A unique solution has been found. The optimized boresight "
			"misalignment angles in stage 2 are:"<<endl
		  <<"   phi   = 180 deg + "<<(phi*RAD_TO_DEG-180.)*3600.<<" arcsec\n"
		  <<"   theta = "<<theta*RAD_TO_DEG*3600.<<" arcsec\n"
		  <<"   psi   = "<<psi*RAD_TO_DEG<<" deg\n\n"
		  <<"The corresponding entry for the CCF Boresight constituent is:\n\n"
		  <<Xmm::toString(instrument)<<' '
		  <<setprecision(12)<<phi
		  <<setprecision(12)<<' '
		  <<setprecision(8)<<theta<<' '<<psi
		  <<" 0. 0. 0.\n"
		  <<Sas::message(Msg::App, Msg::Always);

	errstr<<"You have accomplished a lot today - take the rest of the day off!"
		  <<Sas::message(Msg::App, Msg::Always);

	delete slRef;
} // epicbscalgen()
