// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//
//	insbs: compute INStrument BoreSight from star tracker BS
//
//	USAGE:
//		insbs instrument=<instr> ra=<ra> dec=<dec> apos=<apos>
//			  odffixout=yes|no bstoolsout=yes|no
//
//	CHANGE HISTORY:
//		UL, 2000-09-13: 1.0.0: - first version
//		UL, 2000-11-08: 1.1.0: - fixed SPR 1822: inconsistent apos calculated
//								 by INSBS in different output formats
//		UL, 2003-02-06: 1.2.0: - ra/dec are now parameters of type angle
//
//	AUTHOR:
//		U. Lammers, Wed Sep 13 10:16:45 MET DST 2000
//
//	$Header: /sasbuild/sas_git_repo/cvs_copy/sasdev/sas/packages/ccftools/bstools/src/insbs.cc,v 1.8 2003/02/12 16:07:44 ulammers Exp $
//
#include <iostream>
#include <iomanip>

#include "CalServer.h"
#include "Celestial.h"
#include "OalUtils.h"
#include "param.h"
#include "errstr.h"

#include "insbs.h"

void insbs()
{
	//
	//	read input parameters:
	//		Right Ascension/Declination of STR BS + astronomical position angle
	//
	RightAscension ra(realParameter("ra")*DEG_TO_RAD);
	Declination dec(realParameter("dec")*DEG_TO_RAD);
	double apos = realParameter("apos")*DEG_TO_RAD;

	//
	//	construct S/C attitude matrix from STR boresight (STR) 
	//	+ APA with a 3-2-1 Euler angle sequence
	//
	AttitudeMatrix ASC = AttitudeMatrix(ra.angle()+M_PI, dec.angle(), apos,
										EulerAngles::Euler321);
	errstr<<"S/C attitude matrix:\n"<<ASC<<endl
		  <<Sas::message(Msg::App, Msg::Noisy);

	//
	//	Retrieve boresight misalignment matrix (BSM) of specified instrument
	//	from CCF
	//
	string instrString(stringParameter("instrument"));
	calServer.setInstrument(Xmm::toInstrument(instrString));
	Boresight *bs = 0; bs = calServer.getAtom(bs);
	AttitudeMatrix BSM = bs->matrix(0.);
	errstr<<"Boresight misalignment matrix:\n"<<BSM<<endl
		  <<Sas::message(Msg::App, Msg::Noisy);

	//
	//	The BSM maps vectors from the S/C frame into the instrument BS frame
	//	i.e., the INS frame is defined as:
	//		BSA = BSM * ASC
	//	with
	//		BSA	: direction cosine matrix of INS
	//		BSM : boresight misalignment matrix
	//		ASC : direction cosine matrix of S/C frame
	//
	AttitudeMatrix BSA = BSM * ASC;
	errstr<<"Instrument attitude matrix:\n"<<BSA<<endl
		  <<Sas::message(Msg::App, Msg::Noisy);

	//
	//	Now compute Ra/Dec of INS viewing direction which is simply
	//		BSA^-1 * (1, 0, 0)
	//
	EulerAngles ea = BSA.toEulerAngles(EulerAngles::Euler321);
	CartesianVector insBSv = BSA.inverse() * CartesianVector(1., 0., 0.);
	errstr<<"Instrument boresight: "
		  <<setprecision(12)<<insBSv<<endl<<Sas::message(Msg::App, Msg::Noisy);
	EquatorialDirection insBS(insBSv);
	double psi=-ea.psi();
	if (psi < 0)
		psi += M_TWO_PI;

	if (!booleanParameter("bstoolsout"))
		cout<<"Viewing direction of INS:\n"
			<<insBS.ra()<<" / "<<insBS.dec()<<" APOS= "
			<<setprecision(5)<<psi*RAD_TO_DEG<<"deg\n";
	else {
		cout<<setprecision(8)<<"ra="<<insBS.ra().angle()*RAD_TO_DEG
			<<" dec="<<insBS.dec().angle()*RAD_TO_DEG
			<<" apos="<<psi*RAD_TO_DEG;
	}
}
