// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//
//	srcinfov: compute for each instrument where in the FOV a specified
//			  celestial source will appear
//
//	USAGE:
//		strbs strra=<ra> strdec=<dec> apos=<apos> withsrcoord=true|false
//			  srcra=<ra> srcdec=<dec>
//

#include <iostream>
#include <iomanip>

#include "CalServer.h"
#include "Celestial.h"
#include "errstr.h"
#include "Xmm.h"
#include "param.h"

#include "srcinfov.h"

void srcinfov()
{
	//
	//	construct attitude matrix
	//
	RightAscension ra(realParameter("strra")*DEG_TO_RAD);
	Declination dec(realParameter("strdec")*DEG_TO_RAD);
	double apos = realParameter("apos")*DEG_TO_RAD;
	SpacecraftAttitude SCA(ra.angle(), dec.angle(), apos);

	if (booleanParameter("withsrccoord")) {
		ra = RightAscension(realParameter("srcra")*DEG_TO_RAD);
		dec = Declination(realParameter("srcdec")*DEG_TO_RAD);
	}
	CartesianVector srcVec = EquatorialDirection(ra, dec).toUnitCartesian();
    for (unsigned i = 0; i < xmmPayLoad.instruments().size(); ++i) {
        XmmInstrument *ins = xmmPayLoad.instruments()[i];
        if (ins->id() == Xmm::ERM)
            continue;

		calServer.setInstrument(ins->id());
		Boresight *bs = 0; bs = calServer.getAtom(bs);
		bs->withSacCoordOrientation(SCA);

		XmmTelescopePointList pl = bs->projectOntoTelCoord(srcVec);
		XmmCartesianPointList cpl = pl.toCamCoord2().toCamCoord1();
		vector<Ccd> ccd = cpl.onCcd();
		XmmPixelList ccl = cpl.toChipCoord();
		cout<<setw(5)<<Xmm::toString(ins->id())<<": theta = "
			<<setprecision(5)<<pl.theta()[0]<<"'' phi = "
			<<setprecision(5)<<pl.phi()[0]*RAD_TO_DEG
			<<"deg";
		if (ins->id() != Xmm::RGS1 && ins->id() != Xmm::RGS2) {
			cout<<" - this is ";
			if (ccd[0] != Xmm::CCD_NONE)
				cout<<"on pixel "<<ccl.x()[0]<<'/'<<ccl.y()[0]
					<<" of CCD #"<<ccd[0];
			else
				cout<<"not on any CCD (either outside FOV or in chip gap)";
		}
		cout<<endl;
	}
}
